# FIAIF is an Intelligent firewall, version: $Revision: 1.11 $
#
# description: Automates a packet filtering firewall with iptables.
#
# Script Author: Anders Fugmann <afu at fugmann dot net>
# 
# FIAIF is an Intelligent firewall
# Copyright (C) 2002-2003 Anders Peter Fugmann
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

###############################################################################
# Sanity checks to TCP protocol (portscan) 
# Parameters: <queue> <target> [table]
# This is for new connections only
###############################################################################
function add_sanity_check()
{
    local QUEUE=$1
    local TARGET=$2
    local TABLE=$3
    local PARAM=""
    QUEUE="-A ${QUEUE}"
    if [ -n "${TABLE}" ]; then
	QUEUE="-t ${TABLE} ${QUEUE}"
    fi
    
    # XMAS packets. 
    IPTABLES ${QUEUE} -p tcp --tcp-flags ALL          ALL         -j ${TARGET}

    # NULL packets. 
    IPTABLES ${QUEUE} -p tcp --tcp-flags ALL          NONE        -j ${TARGET}

    # FIN/URG/PSH packets.
    IPTABLES ${QUEUE} -p tcp --tcp-flags ALL          FIN,URG,PSH -j ${TARGET}

    # SYN/RST packets.
    IPTABLES ${QUEUE} -p tcp --tcp-flags SYN,RST      SYN,RST     -j ${TARGET}

    # SYN/FIN scan.
    IPTABLES ${QUEUE} -p tcp --tcp-flags SYN,FIN      SYN,FIN     -j ${TARGET}

    # ACK scan (without SYN, RST or FIN)
    IPTABLES ${QUEUE} -p tcp --tcp-flags SYN,ACK,RST,FIN  ACK     -j ${TARGET}

    # FIN packet without ACK.
    IPTABLES ${QUEUE} -p tcp --tcp-flags FIN,ACK      FIN         -j ${TARGET}

    # Silently drop all new TCP without syn flag.
    IPTABLES ${QUEUE} -p tcp \! --syn -j NOLOG_DROP


    ## Forget it. I will _NOT_ block out bit 6 or 7, just because 
    ## they are serveved. Reserved means that they should not be 
    ## taken into account. Only brain damaged firewalls block them out.
    ## Remember ECN?
}
